/*
  James R. Diebel
  Stanford University
  
  Started: 30 November 2004
  Last revised: 

  compmesh.cc - Utility to compare two MSH files in an OpenGL window

  Depends on:
  - Mesh class (mesh.hh)
  -- Vec3d class (vec3d.hh)
  -- Vert class (vert.hh)
  -- Edge class (edge.hh)
  -- Face class (face.hh)
  - Timer class (timer.hh)
  - TriMesh class (trimesh.h)
  - GUI headers (gui.hh)
  -- OpenGL
  -- GLUT
  -- GLUI
*/

#include <iostream>
#include <fstream>
#include <string.h>
#include "bmtk/mesh.hh"
#include "bmtk/gui.hh"
#include "bmtk/timer.hh"

#define D2R 0.017453293
#define R2D 57.29578

using namespace std;

// Data contained within namespace to avoid collisions
namespace compmesh {
  // Mesh data to be displayed
  Mesh* m;
  Mesh* dm;

  // Windowing details (default values if no options file)
  int wx0 = 200; // upper left of main window
  int wy0 = 0;
  int wdx = 1000; // size of main window
  int wdy = 800;
  int cx0 = 0; // upper left of control window
  int cy0 = 0;
  int bdx, bdy; // size of window manager border

  // Viewpoint variables (default values if no options file)
  float xyAspect;
  int   lastX, lastY;
  float rotationX = 0.0, rotationY = 0.0;
  int motionMode;
  Vec3d center;

  // These are the live variables passed into GLUI
  // (default values if no options file)
  int currMesh = 1;
  int origMesh = 0;
  int wireframe = 0;
  int face2face = 0;
  int face2vert = 0;
  int edgeAngles = 0;
  int vertPotentials = 0;
  int vertGradients = 0;
  int faceColors = 1;
  int centerSphere = 1;
  int meshBoundary = 0;
  int vertGradVecs = 0;
  int segment = 0;
  
  int keypoints = 0;
  int vertNormals = 0;
  int faceAreas = 0;
  int vertSteps = 0;
  int edge2face = 0;
  int face2edge = 0;
  int polyMesh = 0;

  float gs2 = 1.0;
  float gsn2 = 1.0;

  char text[sizeof(GLUI_String)] = {""};
  int diffuseIntensity = 100;
  int ambientIntensity = 50;
  int main_window;
  float scale = 1.0;
  int counter = 0;

  float delta=1.0;

  // Pointers to the windows and some of the controls we'll create
  GLUI *glui;
  GLUI_Spinner *diffuseSpinner, *ambientSpinner, *deltaSpinner,
    *gs2Spinner, *gsn2Spinner;
  GLUI_Rollout *visualsRollout;

  // Miscellaneous global variables
  GLfloat light_ambient[4];
  GLfloat light_diffuse[4];
  GLfloat light_position[] = {.5f, .5f, 1.0f, 0.0f};
}
using namespace compmesh;

int main (int argc, char *argv[]) {
  // if bad command line args, give usage instructions and exit
  if (argc < 3) {
    cout << endl 
	 << "Usage: compmesh file_1.msh file_2.msh" << endl;
    exit(1);
  }  

  // read mesh from file
  Mesh m1,m2;
  m1.readMesh(argv[1]); m2.readMesh(argv[2]);
  
  // find mesh properties that are independent of options
  m1.findMeshProps(); m2.findMeshProps();

  // initialize OpenGL window class
  startGUI(&m1, &m2);
  
  return 0;
}

// GLUI control callback
void controlCB(int control) {
  if (control == -1) {
    cout << "Exiting..." << endl << flush;
    exit(0);
  }
  if (control == 100) {
    saveOptions();
    m->writeOptions("mrf_options.dat");
  }
  if (control == 150) {
    center = (m->boxMin + m->boxMax)/2;
  }
  if (control == 151) {
    m->runCG();
  }
  if (control == 152) {
    m->reset();
    sprintf(text,"Potential: %.2f",m->psi);
    myGlutDisplay();
  }
  if (control == 154) {
    m->anneal();
  }
  if (control == 155) {
    m->runSegmentation();
    update();
  }

  if (control == 156) {
    cout << "Saving Mesh..." << endl;
    m->writeMesh("blah");
  }

  if (control == 157) {
    if (dm) delete dm;
    dm = m->genDecimatedMesh();
    update();
  }

  if (control == 250) {
    for (int i=0;i<3;i++) {
      light_diffuse[i] = float(diffuseIntensity)/100.0f;
      light_ambient[i] = float(ambientIntensity)/100.0f;
    }
    glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
    glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient);
  }
  if (control == 251) {
    m->setVertVar(gs2);
    m->setNormalVar(gsn2);
    m->findMeshPotential();
    m->findLocalEdgePotentials();
    m->findVertGradients();
    m->findSearchDirs(true);
    update();
  }
}

// Parses keyboard commands
void myGlutKeyboard(unsigned char Key, int x, int y) {
  switch(Key)
    {
    case 27: 
    case 'q':
      exit(0);
      break;
    case 'x': case ' ':
      center[0] += 0.01*m->box*cos(rotationX*D2R)*sin(rotationY*D2R);    
      center[1] -= 0.01*m->box*sin(rotationX*D2R);
      center[2] -= 0.01*m->box*cos(rotationX*D2R)*cos(rotationY*D2R);
      break;
    case 'z':
      center[0] -= 0.01*m->box*cos(rotationX*D2R)*sin(rotationY*D2R);    
      center[1] += 0.01*m->box*sin(rotationX*D2R);
      center[2] += 0.01*m->box*cos(rotationX*D2R)*cos(rotationY*D2R);
      break;
    case 'a':
      center[0] -= 0.01*m->box*cos(rotationY*D2R);
      center[2] -= 0.01*m->box*sin(rotationY*D2R);
      break;
    case 'd':
      center[0] += 0.01*m->box*cos(rotationY*D2R);
      center[2] += 0.01*m->box*sin(rotationY*D2R);
      break;
    case 'w':
      center[0] += 0.01*m->box*sin(rotationX*D2R)*sin(rotationY*D2R);
      center[1] += 0.01*m->box*cos(rotationX*D2R);
      center[2] -= 0.01*m->box*sin(rotationX*D2R)*cos(rotationY*D2R);
      break;
    case 's':
      center[0] -= 0.01*m->box*sin(rotationX*D2R)*sin(rotationY*D2R);
      center[1] -= 0.01*m->box*cos(rotationX*D2R);
      center[2] += 0.01*m->box*sin(rotationX*D2R)*cos(rotationY*D2R);
      break;
    };
  
  glutPostRedisplay();
}

// Passes keyboard commands to the appropriate interpreter
void myGlutMenu(int value) {
  myGlutKeyboard(value, 0, 0);
}

// Run each time OpenGL renders a frame
void myGlutIdle(void) {
  // According to the GLUT specification, the current window is 
  // undefined during an idle callback.  So we need to explicitly change
  // it if necessary
  if (glutGetWindow() != main_window) 
    glutSetWindow(main_window);  

  glutPostRedisplay();

  // if it is our first iterations,
  // infer the size of the window manager border
  if (!counter) {
    bdx = glutGet(GLUT_WINDOW_X) - wx0;
    bdy = glutGet(GLUT_WINDOW_Y) - wy0;
  }

  // Updates frame counter
  counter++;
}

// Parses mouse button presses
void myGlutMouse(int button, int button_state, int x, int y) {
  lastX = x;
  lastY = y;
  if (button == GLUT_LEFT_BUTTON && button_state == GLUT_DOWN) {
    if (glutGetModifiers() == GLUT_ACTIVE_CTRL)
      motionMode = 3; // translate on surface of screen
    else
      motionMode = 1; // rotate
  }
  if (button == GLUT_RIGHT_BUTTON && button_state == GLUT_DOWN) {
    if (glutGetModifiers() == GLUT_ACTIVE_CTRL)
      motionMode = 4; // translate into/out of screen
    else 
      motionMode = 2; // zoom
  }

  if (button_state == GLUT_UP) {
    motionMode = 0; // no motion
    myGlutDisplay();
  }
}

// Parses mouse motion
void myGlutMotion(int x, int y) {
  if (motionMode == 1) { // rotate
    rotationX += float(y - lastY)/4.0f;
    rotationY += float(x - lastX)/4.0f;
    //cout << rotationX << " " << rotationY << endl << flush;
  } else if (motionMode == 2) { // zoom
    int diff = y - lastY;
    int mag = abs(y - lastY);
    for (int i=0;i<mag;i++) {
      if (diff < 0) scale *= 1.01;
      else scale /= 1.01;
    }
  } else if (motionMode == 3) { // translate left/right, up/down
    float distX = 0.1*m->box*float(x - lastX)/float(wdx);
    float distY = 0.1*m->box*float(y - lastY)/float(wdy);
    center[0] -= distX*cos(rotationY*D2R);
    center[2] -= distX*sin(rotationY*D2R);
    center[0] += distY*sin(rotationX*D2R)*sin(rotationY*D2R);
    center[1] += distY*cos(rotationX*D2R);
    center[2] -= distY*sin(rotationX*D2R)*cos(rotationY*D2R);
  } else if (motionMode == 4) { // translate into/out of screen
    float distZ = 0.1*m->box*float(y - lastY)/float(wdy);
    center[0] -= distZ*cos(rotationX*D2R)*sin(rotationY*D2R);    
    center[1] += distZ*sin(rotationX*D2R);
    center[2] += distZ*cos(rotationX*D2R)*cos(rotationY*D2R);
  }
  if (motionMode) {
    lastX = x;
    lastY = y;
  }

  glutPostRedisplay(); 
}

// tells GLUT when the window has been reshaped
void myGlutReshape(int x, int y) {
  wdx = x; wdy = y;
  xyAspect = (float)x / (float)y;
  glViewport(0, 0, x, y);

  glutPostRedisplay();
}


float neg(float arg) {
  return (fabs(arg) - arg)/2;
}

// Draws the actual mesh with calls directly to OpenGL
void drawMesh(void) {
  // Draw the triangles
  if (currMesh || origMesh) {
    float color;
    glBegin(GL_TRIANGLES);
    for (int i=0;i<m->nf;i++) {
      float* n;
      if (currMesh) n = &(m->f[i].n.x[0]);
      if (origMesh) n = &(m->f[i].n0.x[0]);
      if (segment) n = &(m->f[i].r->n.x[0]);
      glNormal3d(n[0],n[1],n[2]);
      
      if (faceColors) {
	glColor4f(fabs(n[0]),fabs(n[1]),fabs(n[2]),0.5);
      } else if (faceAreas) {
	color = m->f[i].s/delta;
	glColor4f(color,color,color,0);
      } else glColor4f(0.5,0.5,0.5,0.5);
      for (int j=0;j<3;j++) {
	if (vertPotentials) { 
	  color = (m->f[i].v[j]->psiEdge + m->f[i].v[j]->psi)/delta; 
	  glColor4f(color,color,color,0);
	}
	if (vertGradients) {
	  glColor4f(fabs(m->f[i].v[j]->gradPsi[0])/delta,
		    fabs(m->f[i].v[j]->gradPsi[1])/delta,
		    fabs(m->f[i].v[j]->gradPsi[2])/delta,0);
	}
	if (vertSteps) {
	  color = m->f[i].v[j]->step*delta;
	  glColor4f(color,color,color,0);
	}
	if (segment && 0) {
	  if (m->f[i].flag < -1) {
	    color = 1;
	    glColor4f(color,color,color,0);
	  }
	}
	if (!origMesh)
	  glVertex3d(m->f[i].v[j]->x[0], m->f[i].v[j]->x[1], m->f[i].v[j]->x[2]);
	else 
	  glVertex3d(m->f[i].v[j]->x0[0], 
		     m->f[i].v[j]->x0[1], 
		     m->f[i].v[j]->x0[2]);
      }
    }
    glEnd();
  }

  // Draw the triangles
  if (segment && dm) {
    float color;
    glBegin(GL_TRIANGLES);
    for (int i=0;i<dm->nf;i++) {
      float* n;
      n = &(dm->f[i].n.x[0]);
      glNormal3d(n[0],n[1],n[2]);
      
      if (faceColors) {
	glColor4f(n[0],n[1],n[2],0.5);
      } else glColor4f(0.5,0.5,0.5,0.5);
      for (int j=0;j<3;j++) {
	glVertex3d(dm->f[i].v[j]->x[0], 
		   dm->f[i].v[j]->x[1], 
		   dm->f[i].v[j]->x[2]);
      }
    }
    glEnd();
  }
  // Draw the triangles
  if (keypoints && dm) {
    float color;
    glBegin(GL_POINTS);
    for (int i=0;i<dm->nf;i++) {
      glColor4f(1,1,1,0.5);
      for (int j=0;j<4;j++) {
	if (dm->f[i].v[j%3]->i == 141) 
	  glVertex3d(dm->f[i].v[j%3]->x[0], 
		     dm->f[i].v[j%3]->x[1], 
		     dm->f[i].v[j%3]->x[2]);
      }
    }
    glEnd();
  }

  // Draw the wireframe
  if (wireframe) {
    glBegin(GL_LINES);
    if (!edgeAngles) glColor4f(0,0,0,0.5);
    glNormal3d(0,0,0);
    for (int i=0;i<m->ne;i++) {
      if (segment) {
	if (m->e[i].flag == -1)
	  continue;
      }
      if (edgeAngles) {
	float color = m->e[i].psi/delta;
	glColor4f(color,color,color,0);
      }
      if (!origMesh) {
	glVertex3d(m->e[i].v[0]->x[0],
		   m->e[i].v[0]->x[1],
		   m->e[i].v[0]->x[2]);
	glVertex3d(m->e[i].v[1]->x[0],
		   m->e[i].v[1]->x[1],
		   m->e[i].v[1]->x[2]);
      } else {
	glVertex3d(m->e[i].v[0]->x0[0],
		   m->e[i].v[0]->x0[1],
		   m->e[i].v[0]->x0[2]);
	glVertex3d(m->e[i].v[1]->x0[0],
		   m->e[i].v[1]->x0[1],
		   m->e[i].v[1]->x0[2]);
      }
    }
    glEnd();
  }

  // draw points flagged as boundary points between regions
  if (keypoints) {
    glBegin(GL_POINTS);
    glColor4f(1,1,1,0.5);
    for (int i=0;i<m->nv;i++) {
      if (m->v[i].flag < -3) {
	glVertex3d(m->v[i].x[0],
		   m->v[i].x[1],
		   m->v[i].x[2]);
      }
    }
    glEnd();
  }

  if (segment && 0) {
    glBegin(GL_POINTS);
    glColor4f(1,1,1,0.5);
    for (int i=0;i<m->nr;i++) {
      if (m->r[i].vi.size() > 0) {
	glVertex3d(m->v[m->r[i].vi[0]].x[0],
		   m->v[m->r[i].vi[0]].x[1],
		   m->v[m->r[i].vi[0]].x[2]);
	glVertex3d(m->v[m->r[i].vi.back()].x[0],
		   m->v[m->r[i].vi.back()].x[1],
		   m->v[m->r[i].vi.back()].x[2]);
      }
    }
    glEnd();
  }
    
  if (segment) {
    for (int i=0;i<m->nr;i++) {
      if (i>=m->nr) break;
      glBegin(GL_LINE_STRIP);
      float color = 0;
      glColor4f(color,color,color,0.5f);
      for(int j=0;j<m->r[i].vi.size();j++) {
	if (m->r[i].vi[j] == -1) {
	  glEnd();
	  glBegin(GL_LINE_STRIP);
	  continue;
	} else {
        glVertex3d(m->v[m->r[i].vi[j]].x[0],
                   m->v[m->r[i].vi[j]].x[1],
                   m->v[m->r[i].vi[j]].x[2]);
	}
      }
      glEnd();
    }
  }

  if (polyMesh) {
    for (int i=0;i<m->nr;i++) {
      glBegin(GL_POLYGON);
      if (faceColors) 
	glColor4f(m->r[i].n.x[0],m->r[i].n.x[1],m->r[i].n.x[2],0.5);
      else
	glColor4f(0.50,0.50,0.50,0.5);
      glNormal3d(m->r[i].n.x[0],m->r[i].n.x[1],m->r[i].n.x[2]);
      for(int j=0;j<m->r[i].vi.size();j++) {
	if (m->r[i].vi[j] == -1)
	  break;
	else
	  glVertex3d(m->v[m->r[i].vi[j]].x[0],
		     m->v[m->r[i].vi[j]].x[1],
		     m->v[m->r[i].vi[j]].x[2]);
      }
      glEnd();
    }
  }

  // Draw the mesh boundary
  if (meshBoundary) {
    glBegin(GL_LINES);
    glColor4f(0,0,0,0.5);
    glNormal3d(0,0,0);
    for (int i=0;i<m->ne;i++) {
      if (m->e[i].nf == 1) {
	glVertex3d(m->e[i].v[0]->x[0],
		   m->e[i].v[0]->x[1],
		   m->e[i].v[0]->x[2]);
	glVertex3d(m->e[i].v[1]->x[0],
		   m->e[i].v[1]->x[1],
		   m->e[i].v[1]->x[2]);
      }
    }
    glEnd();

    glBegin(GL_POINTS);
    for (int i=0;i<m->nv;i++) {
      if (m->v[i].bound) {
	glVertex3d(m->v[i].x[0],
		   m->v[i].x[1],
		   m->v[i].x[2]);
      }
    }
    glEnd();

    glBegin(GL_TRIANGLES);
    glColor4f(1,1,1,0.5);
    for (int i=0;i<m->nf;i++) {
      if (m->f[i].bound) {
	glVertex3d(m->f[i].v[0]->x[0],
		   m->f[i].v[0]->x[1],
		   m->f[i].v[0]->x[2]);
	glVertex3d(m->f[i].v[1]->x[0],
		   m->f[i].v[1]->x[1],
		   m->f[i].v[1]->x[2]);
	glVertex3d(m->f[i].v[2]->x[0],
		   m->f[i].v[2]->x[1],
		   m->f[i].v[2]->x[2]);
      }
    }
    glEnd();
  }

  // Draw the gradient vector field
  if (vertGradVecs) {
    glBegin(GL_LINES);
    glColor4f(0,0,0,0.5);
    glNormal3d(0,0,0);
    for (int i=0;i<m->nv;i++) {
      glVertex3d(m->v[i].x[0],m->v[i].x[1],m->v[i].x[2]);
      glVertex3d(m->v[i].x[0]+m->v[i].gradPsi[0]/(2*delta),
		 m->v[i].x[1]+m->v[i].gradPsi[1]/(2*delta),
		 m->v[i].x[2]+m->v[i].gradPsi[2]/(2*delta));
    }
    glEnd();
  }

  // Draw the vertex normal vector field
  if (vertNormals) {
    glBegin(GL_LINES);
    glColor4f(0,0,0,0.5);
    glNormal3d(0,0,0);
    for (int i=0;i<m->nv;i++) {
      glVertex3d(m->v[i].x[0],m->v[i].x[1],m->v[i].x[2]);
      glVertex3d(m->v[i].x[0]+m->v[i].n[0]/delta,
		 m->v[i].x[1]+m->v[i].n[1]/delta,
		 m->v[i].x[2]+m->v[i].n[2]/delta);
    }
    glEnd();
  }

  // Draw the face-to-face links
  if (face2face) {
    glBegin(GL_LINES);
    glColor4f(1.0,0,0,0);
    glNormal3d(0,0,0);
    for (int i=0;i<m->nf;i++) {
      for (int j=0;j<m->f[i].nf;j++) {
	glVertex3d(m->f[i].x[0],m->f[i].x[1],m->f[i].x[2]);
	glVertex3d(m->f[i].f[j]->x[0],m->f[i].f[j]->x[1],m->f[i].f[j]->x[2]);
      }
    }
    glEnd();
  }

  // Draw the face-to-vertex links
  if (face2vert) {
    glBegin(GL_LINES);
    glColor4f(0,1.0,0,0);
    glNormal3d(0,0,0);
    for (int i=0;i<m->nf;i++) {
      for (int j=0;j<3;j++) {
	glVertex3d(m->f[i].x[0],m->f[i].x[1],m->f[i].x[2]);
	glVertex3d(m->f[i].v[j]->x[0],m->f[i].v[j]->x[1],m->f[i].v[j]->x[2]);
      }
    }
    glEnd();
  }

  // Draw edge-to-face links
  if (edge2face) {
    glBegin(GL_LINES);
    glColor4f(1.0,0,0,0);
    glNormal3d(0,0,0);
    for (int i=0;i<m->ne;i++) {
      for (int j=0;j<m->e[i].nf;j++) {
	glVertex3d(0.5*(m->e[i].v[0]->x[0] + m->e[i].v[1]->x[0]),
		   0.5*(m->e[i].v[0]->x[1] + m->e[i].v[1]->x[1]),
		   0.5*(m->e[i].v[0]->x[2] + m->e[i].v[1]->x[2]));
	glVertex3d(m->e[i].f[j]->x[0],m->e[i].f[j]->x[1],m->e[i].f[j]->x[2]);
      }
    }
    glEnd();
  }

  // Draw face-to-edge links
  if (face2edge) {
    glBegin(GL_LINES);
    glColor4f(1.0,0,0,0);
    glNormal3d(0,0,0);
    for (int i=0;i<m->nf;i++) {
      for (int j=0;j<3;j++) {
	glVertex3d(m->f[i].x[0],m->f[i].x[1],m->f[i].x[2]);
	glVertex3d(0.5*(m->f[i].e[j]->v[0]->x[0] + m->f[i].e[j]->v[1]->x[0]),
		   0.5*(m->f[i].e[j]->v[0]->x[1] + m->f[i].e[j]->v[1]->x[1]),
		   0.5*(m->f[i].e[j]->v[0]->x[2] + m->f[i].e[j]->v[1]->x[2]));
      }
    }
    glEnd();
  }

}


// Displays everything to the screen
void myGlutDisplay(void) {
  //glClearColor(.9f, .9f, .9f, 1.0f);
  glClearColor(0.3, 0.3, 0.3, 1.0f);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glFrustum(-xyAspect*.0008, xyAspect*.0008, -.0008, .0008, .01, 15.0);
  glMatrixMode(GL_MODELVIEW);

  glLoadIdentity();
  glTranslatef(0.0, 0.0, -1.2f);
  glRotatef(rotationX, 1.0, 0.0, 0.0);
  glRotatef(rotationY, 0.0, 1.0, 0.0);

  // Render the center of rotation
  if (centerSphere || motionMode > 2) {
    glColor4f(1,0,0,0);
    glutSolidSphere(0.005,10,10);
  }

  // Scale things according to zoom factor
  glScalef(scale, scale, scale);

  // Render the mesh
  glTranslatef(-center[0], -center[1], -center[2]);
  drawMesh();
  glTranslatef(center[0], center[1], center[2]);

  // Disable lighting and set up ortho projection to render text
  glDisable(GL_LIGHTING);  
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluOrtho2D(0.0, 100.0, 0.0, 100.0);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glColor3ub(255,255,255);
  glRasterPos2i(10, 10);

  // Render the live character array 'text'
  for(int i=0;i<(int)strlen(text);i++)
    glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18, text[i]);

  glEnable(GL_LIGHTING);

  glutSwapBuffers(); 
}

void saveOptions() {
  cout << "- Saving GUI options..." << flush;
  ofstream fout("gui_options.dat");
  if (fout) {
    // Window details
    fout << glutGet(GLUT_WINDOW_X) - bdx << " ";
    fout << glutGet(GLUT_WINDOW_Y) - bdy << endl;
    fout << glutGet(GLUT_WINDOW_WIDTH) << " ";
    fout << glutGet(GLUT_WINDOW_HEIGHT) << endl;
    fout << cx0 << endl;
    fout << cy0 << endl;

    // Viewpoint variables
    fout << center[0] << " ";
    fout << center[1] << " ";
    fout << center[2] << endl;
    fout << rotationX << " ";
    fout << rotationY << endl;
    fout << scale << endl;
    fout << delta << endl;

    // Visuals
    fout << currMesh << endl;
    fout << origMesh << endl;
    fout << wireframe << endl;
    fout << face2face << endl;
    fout << face2vert << endl;
    fout << edgeAngles << endl;
    fout << vertPotentials << endl;
    fout << vertGradients << endl;
    fout << faceColors << endl;
    fout << centerSphere << endl;
    fout << meshBoundary << endl;
    fout << vertGradVecs << endl;
    fout << segment << endl;
    fout << diffuseIntensity << endl;
    fout << ambientIntensity << endl;
  }

  fout.close();
  cout << "Done." << endl << flush;
}

void loadOptions() {
  cout << "- Loading GUI options..." << flush;
  ifstream fin("gui_options.dat");
  if (fin) {
    // Window details
    fin >> wx0;
    fin >> wy0;
    fin >> wdx;
    fin >> wdy;
    fin >> cx0;
    fin >> cy0;

    // Viewpoint variables
    fin >> center[0];
    fin >> center[1];
    fin >> center[2];
    fin >> rotationX;
    fin >> rotationY;
    fin >> scale;
    fin >> delta;

    // Visuals
    fin >> currMesh;
    fin >> origMesh;
    fin >> wireframe;
    fin >> face2face;
    fin >> face2vert;
    fin >> edgeAngles;
    fin >> vertPotentials;
    fin >> vertGradients;
    fin >> faceColors;
    fin >> centerSphere;
    fin >> meshBoundary;
    fin >> vertGradVecs;
    fin >> segment;
    fin >> diffuseIntensity;
    fin >> ambientIntensity;
  }

  for (int i=0;i<3;i++) {
    light_diffuse[i] = float(diffuseIntensity)/100.0f;
    light_ambient[i] = float(ambientIntensity)/100.0f;
  } 
  gs2 = m->gs2;
  gsn2 = m->gsn2;
  fin.close();
  cout << "Done." << endl << flush;
}

void update(void) {
  sprintf(text,"Potential: %.2f",m->psi);
  myGlutDisplay();
}

// Starts the graphical user interface
int startGUI(Mesh *m_) {
  // Set local variables to passed values
  m = m_;
  sprintf(text,"Potential: %.2f",m->psi);
  m->registerCB(update);

  // Load options file
  loadOptions();

  ///////////////////////////////////////
  // Initialize GLUT and create window //
  ///////////////////////////////////////
  glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
  glutInitWindowPosition(wx0, wy0);
  glutInitWindowSize(wdx, wdy);
 
  main_window = glutCreateWindow("ViewMesh 1.0");
  glutDisplayFunc(myGlutDisplay);
  glutReshapeFunc(myGlutReshape);  
  glutKeyboardFunc(myGlutKeyboard);
  glutMotionFunc(myGlutMotion);
  glutMouseFunc(myGlutMouse);

  // Set up OpenGL lights
  glEnable(GL_LIGHTING);
  glEnable(GL_NORMALIZE);

  glEnable(GL_LIGHT0);
  glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
  glLightfv(GL_LIGHT0, GL_POSITION, light_position);

  // Set up OpenGL materials
  glEnable(GL_COLOR_MATERIAL);

  // Make sure lines appear above polygons
  //glPolygonOffset(10.0,10.0);
  //glEnable(GL_POLYGON_OFFSET_FILL);
  glDisable(GL_POLYGON_OFFSET_FILL);

  // Enable z-buferring and others
  glEnable(GL_DEPTH_TEST);
  glClearDepth(1.0f);
  glDepthFunc(GL_LEQUAL);
  glPointSize(4);
  //glEnable(GL_CULL_FACE);
  

  //////////////////////////
  // Here's the GLUI code //
  //////////////////////////
  cout << "- Starting GLUI version " << GLUI_Master.get_version() << endl;

  // name, flags, x, and y
  glui = GLUI_Master.create_glui("Control Panel", cx0, cy0, 0); 

  glui->add_statictext("ViewMesh 1.0"); 
  visualsRollout = glui->add_rollout("Visuals", true);

  // Control for the visual display
  glui->add_checkbox_to_panel
    (visualsRollout,"Current Mesh",&currMesh,0, NULL);
  glui->add_checkbox_to_panel
    (visualsRollout,"Original Mesh",&origMesh,0, NULL);
  glui->add_checkbox_to_panel
    (visualsRollout,"Wireframe",&wireframe,0, NULL);
  glui->add_checkbox_to_panel
    (visualsRollout,"Face-to-Face",&face2face,0, NULL);
  glui->add_checkbox_to_panel
    (visualsRollout,"Face-to-Vertex",&face2vert,0, NULL);
  glui->add_checkbox_to_panel
    (visualsRollout,"Edge Angles",&edgeAngles,0, NULL);
  glui->add_checkbox_to_panel
    (visualsRollout,"Vert Potentials",&vertPotentials,0, NULL);
  glui->add_checkbox_to_panel
    (visualsRollout,"Vert Gradients",&vertGradients,0, NULL);
  glui->add_checkbox_to_panel
    (visualsRollout,"Face Colors",&faceColors,0, NULL);
  glui->add_checkbox_to_panel
    (visualsRollout,"Center Sphere",&centerSphere,0, NULL);
  glui->add_checkbox_to_panel
    (visualsRollout,"Mesh Boundary",&meshBoundary,0, NULL);
  glui->add_checkbox_to_panel
    (visualsRollout,"Vert Normals",&vertNormals,0, NULL);
  glui->add_checkbox_to_panel
    (visualsRollout,"Vert Grad Vecs",&vertGradVecs,0, NULL);
  glui->add_checkbox_to_panel
    (visualsRollout,"Face Areas",&faceAreas,0, NULL);
  glui->add_checkbox_to_panel
    (visualsRollout,"Vert Steps",&vertSteps,0, NULL);
  glui->add_checkbox_to_panel
    (visualsRollout,"Segmentation",&segment,0, NULL);
  glui->add_checkbox_to_panel
    (visualsRollout,"Key Points",&keypoints,0, NULL);
  glui->add_checkbox_to_panel
    (visualsRollout,"PolyMesh",&polyMesh,0, NULL);

  // Add some controls for lights
  diffuseSpinner = glui->add_spinner_to_panel
    (visualsRollout, "Diffuse:", GLUI_SPINNER_INT,
     &diffuseIntensity,250,controlCB);
  diffuseSpinner->set_int_limits(0, 100);
  ambientSpinner = glui->add_spinner_to_panel
    (visualsRollout, "Ambient:", GLUI_SPINNER_INT,
     &ambientIntensity,250,controlCB);
  ambientSpinner->set_int_limits(0, 100);

  deltaSpinner = glui->add_spinner_to_panel
    (visualsRollout, "Delta:", GLUI_SPINNER_FLOAT,
     &delta,0,controlCB);
  deltaSpinner->set_float_limits(0,100);

  gs2Spinner = glui->add_spinner_to_panel
    (visualsRollout, "Vert Var:", GLUI_SPINNER_FLOAT,
     &gs2,251,controlCB);
  gs2Spinner->set_float_limits(0,1000);

  gsn2Spinner = glui->add_spinner_to_panel
    (visualsRollout, "Normal Var:", GLUI_SPINNER_FLOAT,
     &gsn2,251,controlCB);
  gsn2Spinner->set_float_limits(0,2);

  // Recenter button
  glui->add_button("Recenter",150,controlCB);
  // A 'run' button
  glui->add_button("Run",151,controlCB);
  glui->add_button("Run Anneal",154,controlCB);
  glui->add_button("Reset",152,controlCB);
  glui->add_button("Segment",155,controlCB);
  glui->add_button("Tessellate",157,controlCB);

  // Save mesh button
  glui->add_button("Save Mesh", 156,controlCB);

  // Save options button
  glui->add_button("Save Options",100,controlCB);
  // A 'quit' button
  glui->add_button("Quit", -1,controlCB);

  // Link windows to GLUI, and register idle callback
  glui->set_main_gfx_window(main_window);

  // We register the idle callback with GLUI, not with GLUT
  GLUI_Master.set_glutIdleFunc(myGlutIdle);

  // Regular GLUT main loop
  glutMainLoop();

  return 0;
}
