/*
  James R. Diebel
  Stanford University

  Started: 24 August 2004
  Last revised:

  mesh_vert.cc - class implementation of Mesh class (mesh.hh)

  Depends on:
  - Mesh class (mesh.hh)
*/

#include "bmtk/mesh.hh"
#include <iostream>

namespace bmtk {

//////////////////////////
// Per-Vertex Functions //
//////////////////////////
void Mesh::resetVertFlags() {
  if (po) cout << "- Reseting vertex flags..." << flush;
  for (int i=0;i<nv;i++) v[i].flag = v[i].marker = -1;
  if (po) cout << "Done." << endl << flush;
}

void Mesh::findVertNormals() {
  if (po) cout << "- Computing vertex normals..." << flush;
  for (int i=0;i<nv;i++) v[i].findNormal();
  if (po) cout << "Done." << endl << flush;
}

void Mesh::findVertLengths() {
  if (po) cout << "- Computing vertex lengths..." << flush;
  for (int i=0;i<nv;i++) v[i].findLength();
  if (po) cout << "Done." << endl << flush;
}

void Mesh::findVertProps() {
  if (po) cout << "- Computing vertex properties..." << flush;
  for (int i=0;i<nv;i++) {
    v[i].findNormal();
    v[i].findLength();
  }
  if (po) cout << "Done." << endl << flush;
}

void Mesh::findVertDepths() {
  if (po) cout << "- Computing vertex depths..." << flush;
  for (int i=0;i<nv;i++) {
    v[i].findDepth();
  }
  if (po) cout << "Done." << endl << flush;
}

void Mesh::findVertPotentials() {
  if (po) cout << "- Computing vertex potentials..." << flush;
  for (int i=0;i<nv;i++) {
    v[i].findPotential();
  }
  if (po) cout << "Done." << endl << flush;
}

void Mesh::findLocalEdgePotentials() {
  if (po) cout << "- Computing local edge potentials..." << flush;
  bool temp = po;
  po = false;
  resetEdgeFlags();
  po = temp;
  for (int i=0;i<nv;i++) {
    v[i].findLocalEdgePotential();
  }
  if (po) cout << "Done." << endl << flush;
}

void Mesh::findVertGradients() {
  if (ut) tVG->start();
  if (po) cout << "- Computing local potential gradients..." << flush;
  bool temp = po;
  po = false;
  resetEdgeFlags();
  if (temp) cout << " edge flags, " << flush;
  for (int i=0;i<nv;i++) {
    v[i].findGradient();
  }
  po = temp;
  if (po) cout << "Done." << flush;
  if (po) tVG->printMark();
  if (ut) tVG->mark();
}

void Mesh::findNormGradPsi() {
  if (po)
    cout << "- Computing magnitude of gradient along search direction..."
   << flush;
  normGradPsi = 0;
  for (int i=0;i<nv;i++) {
    normGradPsi += v[i].gradPsi.dot(v[i].dir);
  }
  if (po) cout << "Done." << endl << flush;
}

void Mesh::findSearchDirs(bool reset) {
  if (po) cout << "- Computing search direction..." << flush;
  if (po) if (reset) cout << " with CG restart... " << flush;
  normDir = 0;
  rho[iter%2] = 0;
  for (int i=0;i<nv;i++) {
    rho[iter%2] += v[i].gradPsi.len2();
    v[i].dir0 = v[i].dir;
  }
  if (reset) {
    for (int i=0;i<nv;i++) {
      v[i].dir = -v[i].gradPsi;
      normDir += ~(v[i].dir);
    }
  } else {
    float beta = rho[iter%2]/rho[(iter-1)%2];
    if (po) cout << " beta = " << beta << ", " << flush;
    for (int i=0;i<nv;i++) {
      v[i].findSearchDir(beta);
      normDir += ~(v[i].dir);
    }
  }
  if (po) cout << "|d| = " << normDir/nv << ", Done." << endl << flush;
}

void Mesh::saveRefVerts() {
  if (po) cout << "- Saving reference vertex positions..." << flush;
  for (int i=0;i<nv;i++) v[i].xr = v[i].x;
  if (po) cout << "Done." << endl << flush;
}

void Mesh::moveVerts(float step) {
  if (po) cout << "- Moving vertices with global step..." << flush;
  for (int i=0;i<nv;i++) v[i].move(step);
  if (po) cout << "Done." << endl << flush;
}

void Mesh::moveVerts() {
  if (po) cout << "- Moving vertices with local steps..." << flush;
  for (int i=0;i<nv;i++) v[i].move();
  if (po) cout << "Done." << endl << flush;
}

void Mesh::vertFaceConsistency(int num) {
  if (po) cout << "- Moving toward vertex-normal consistency..." << flush;
  bool temp = po;
  po = false;
  for (int i=0;i<num;i++) {
    saveRefVerts();
    for (int i=0;i<nv;i++) v[i].flatten();
    findFaceCenters();
  }
  po = temp;
  if (po) cout << "Done." << endl << flush;
}

} // namespace bmtk
